package devmgr.api.melutil;

import devmgr.api.symbol.ChannelLocation;
import devmgr.api.symbol.ChannelType;
import devmgr.api.symbol.ComponentLocation;
import devmgr.api.symbol.ComponentRelativeLocation;
import devmgr.api.symbol.EventComponentType;
import devmgr.api.symbol.EventPriority;
import devmgr.api.symbol.ExtendedComponentLocation;
import devmgr.api.symbol.ExtendedComponentLocationType;
import devmgr.api.symbol.IpAddressType;
import devmgr.api.symbol.IpVxAddress;
import devmgr.api.symbol.MelEntry;
import devmgr.api.symbol.MinihubLocation;
import devmgr.api.symbol.MinihubType;
import devmgr.api.symbol.ParentRelativeLocation;
import devmgr.api.symbol.TraySlot;

import java.text.MessageFormat;
import java.util.Locale;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

/**
DESCRIPTION:                                                            <BR>
The EventUtil class provides various methods to "decode" the various fields contained
in a melEntry. It "decodes" the fields to string format for use by any other class.
These methods were pulled (taken out of) from the lsi.symsm.sam.eventviewer.EventDecoder
class to allow other classes to access them. Initially, the SAStatusMonitor class
needed these methods - getDescription and getUSASCIIDescription.
<BR><BR>

NOTES/ASSUMPTIONS:                                                      <BR>
This class should not be instantiated. Other classes should just call the static
methods directly.
<BR><BR>

REFERENCE:                                                              <BR>
User Interface Design - Event Viewer (349-1001830)
Design Detail - Event Viewer (349-1002090)
Feature Approach Document - Event Viewer (349-1001540)
*/

public class EventUtil
{
   /**
   DESCRIPTION: <BR>
   The EventUtil class contains all public static members and methods.
   */

   private static final String EVENT_DESCRIPTION_KEY_PREFIX = "EV_DESCRIPTION_";

   private static ResourceBundle    m_USASCIIResource = null;
   private static final int         LINE_LENGTH = 16;
   private static final int         CONTROLLER_TRAY = -1;
   private static final int         TRAY_LOC_UNKNOWN = 100;
   
   private static final int         SLOT_A = 1;
   private static final int         SLOT_B = 2;
                                         
   private static String []   m_Categories;
   
   /* Note: The order of these entries MUST match up to the order of the corresponding
    * entries in the following array (m_ComponentTypeEnums).
    */
   private static  String []   m_ComponentTypes;
   static String[] getComponentTypes(){
        return m_ComponentTypes;
    }

   private static EventComponentType []      m_ComponentTypeEnums;
   static EventComponentType[] getComponentTypeEnum(){
        return m_ComponentTypeEnums;
    }


   public static void init()
   {
	   initCategories();
	   initComponentType();
   }

   private static void initCategories()
   {
	   m_Categories = new String [] { 
                                       Resource.getString("EV_CATEGORY_UNKNOWN"),
                                       Resource.getString("EV_CATEGORY_ERROR"),
                                       Resource.getString("EV_CATEGORY_FAILURE"),
                                       Resource.getString("EV_CATEGORY_COMMAND"),
                                       Resource.getString("EV_CATEGORY_NOTIFICATION"),
                                       Resource.getString("EV_CATEGORY_STATE_CHANGE"),
                                       Resource.getString("EV_CATEGORY_HOST_ENTRY"),
                                       Resource.getString("EV_CATEGORY_GENERAL")};
   }
   
   private static void initComponentType()
   {
	   m_ComponentTypes = new String [] { 
                                       Resource.getString("EV_COMPONENT_TYPE_UNKNOWN"),
                                       Resource.getString("EV_COMPONENT_TYPE_DRIVE"),
                                       Resource.getString("EV_COMPONENT_TYPE_CONTROLLER"),
                                       Resource.getString("EV_COMPONENT_TYPE_BATTERY"),
                                       Resource.getString("EV_COMPONENT_TYPE_FAN"),
                                       Resource.getString("EV_COMPONENT_TYPE_MINI_HUB"),
                                       Resource.getString("EV_COMPONENT_TYPE_POWER_SUPPLY"),
                                       Resource.getString("EV_COMPONENT_TYPE_ENCLOSURE"),
                                       Resource.getString("EV_COMPONENT_TYPE_ENV_CARD"),
                                       Resource.getString("EV_COMPONENT_TYPE_UPS"),
                                       Resource.getString("EV_COMPONENT_TYPE_CHIP"),
                                       Resource.getString("EV_COMPONENT_TYPE_VOLUME"),
                                       Resource.getString("EV_COMPONENT_TYPE_VOL_GROUP"),
                                       Resource.getString("EV_COMPONENT_TYPE_TEMP_SENSOR"),
                                       Resource.getString("EV_COMPONENT_TYPE_PORT_CRU"),
                                       Resource.getString("EV_COMPONENT_TYPE_CHANNEL"),
                                       Resource.getString("EV_COMPONENT_TYPE_INTERCONNECT_CRU"),
                                       Resource.getString("EV_COMPONENT_TYPE_SUPPORT_CRU"),
                                       Resource.getString("EV_COMPONENT_TYPE_ALARM"),
                                       Resource.getString("EV_COMPONENT_TYPE_CHANNEL_PORT"),
                                       Resource.getString("EV_COMPONENT_TYPE_SFP_PORT"),
                                       Resource.getString("EV_COMPONENT_TYPE_HOST_BOARD"),
                                       Resource.getString("EV_COMPONENT_TYPE_NEW_FORMAT"),
                                       Resource.getString("EV_COMPONENT_TYPE_CTLR_SFP"),
                                       Resource.getString("EV_COMPONENT_TYPE_CTLR_SOC"),
                                       Resource.getString("EV_COMPONENT_TYPE_INITIATOR"),
                                       Resource.getString("EV_COMPONENT_TYPE_TARGET"),
                                       Resource.getString("EV_COMPONENT_TYPE_ISNS_SERVER"),
                                       Resource.getString("EV_COMPONENT_TYPE_HOST_IO_CARD"),
                                       Resource.getString("EV_COMPONENT_TYPE_CACHE_BACKUP_DEVICE"),
                                       Resource.getString("EV_COMPONENT_TYPE_DRAWER"),
                                       Resource.getString("EV_COMPONENT_TYPE_RELATIVE"),
                                       Resource.getString("EV_COMPONENT_TYPE_SCHEDULE"),
                                       Resource.getString("EV_COMPONENT_TYPE_ASYNC_MIRROR_GROUP"),
                                       Resource.getString("EV_COMPONENT_TYPE_DISK_POOL"),
                                       Resource.getString("EV_COMPONENT_TYPE_PIT"),
                                       Resource.getString("EV_COMPONENT_TYPE_PIT_CONSISTENCY_GROUP"),
                                       Resource.getString("EV_COMPONENT_TYPE_CGPIT"),
                                       Resource.getString("EV_COMPONENT_TYPE_CGVIEW"),
                                       Resource.getString("EV_COMPONENT_TYPE_FLASH_CACHE")
                                       };
   
   /*
    * The m_ComponentTypeEnums array is used as a lookup table for m_ComponentTypes.
    * This protects the array m_ComponentTypes from any changes in the enumeration of
    * EventComponentType.  Note that if a new string is added to m_ComponentTypes, a corresponding
    * entry is needed in m_ComponentTypeEnums.
    */
       // todo: make new enum at symbol layer (or here) for EventComponentType and description
	   m_ComponentTypeEnums = new EventComponentType [] {
                                       EventComponentType.unknown,
                                       EventComponentType.drive,
                                       EventComponentType.controller,
                                       EventComponentType.battery,
                                       EventComponentType.fan,
                                       EventComponentType.minihub,
                                       EventComponentType.powerSply,
                                       EventComponentType.enclosure,
                                       EventComponentType.esm,
                                       EventComponentType.ups,
                                       EventComponentType.chip,
                                       EventComponentType.volume,
                                       EventComponentType.volumeGrp,
                                       EventComponentType.tempSensor,
                                       EventComponentType.portCru,
                                       EventComponentType.channel,
                                       EventComponentType.interconnectCru,
                                       EventComponentType.supportCru,
                                       EventComponentType.alarm,
                                       EventComponentType.channelPort,
                                       EventComponentType.sfpPort,
                                       EventComponentType.hostBoard,
                                       EventComponentType.newFormat,
                                       EventComponentType.ctlrSfp,
                                       EventComponentType.ctlrSoc,
                                       EventComponentType.initiator,
                                       EventComponentType.target,
                                       EventComponentType.isnsServer,
                                       EventComponentType.hostIoCard,
                                       EventComponentType.cacheBackupDevice,
                                       EventComponentType.drawer,
                                       EventComponentType.relative,
                                       EventComponentType.schedule,
                                       EventComponentType.asyncMirrorGroup,
                                       EventComponentType.diskPool,
                                       EventComponentType.pit,
                                       EventComponentType.pitConsistencyGroup,
                                       EventComponentType.cgpit,
                                       EventComponentType.cgview,
                                       EventComponentType.flashCache
                                       };
   }
   
   /**
   DESCRIPTION: <BR>
   Converts an integer value to the corresponding string representation of the
   event priority.
   <BR><BR>
   
   @param event      The event to request the priority from.
   @return           String representation of the event priority.
   */
   public static String getEventPriorityString(MelEntry event)
   {
      String returnString = Resource.getString("EV_PRIORITY_UNKNOWN");
      EventPriority value = getEventPriority(event);

      switch (value)
      {
          case priorityCritical:
            returnString = Resource.getString("EV_PRIORITY_CRITICAL");
         break;
         
         case priorityInfo:
            returnString = Resource.getString("EV_PRIORITY_INFO");
         break;
         
         case priorityWarning:
            returnString = Resource.getString("EV_PRIORITY_WARNING");
         break;
         
         case priorityDebug:
            returnString = Resource.getString("EV_PRIORITY_DEBUG");
         break;

         default:
             returnString = Resource.getString("EV_PRIORITY_DEFAULT");
            
      }
      return returnString;
   }
   
   /**
    * Determines the EventPriority for a MELEntry.
    * <P>
    * NOTES/ASSUMPTIONS:  <BR>
    * An OEM may choose to override the default priority.
    *
    * @param event to request the priority from.
    * @return priority for the event.
    */
   public static EventPriority getEventPriority(MelEntry event)
   {
      EventPriority  priority = event.getControlParams().getOemPriority();
      if(priority == EventPriority.priorityDefault)
      {
         priority = event.getPriority();
      }         
      
      return priority;
   }
   
   public static String getControllerSlotName(int slot)
   {

      if (slot == SLOT_A)
      {
          return Resource.getString("CTL_A_LABEL");  
      }
      else if (slot == SLOT_B)
      {
    	  return Resource.getString("CTL_B_LABEL");  
      }
      return "";
   }

   /**
   DESCRIPTION: <BR>
   Converts an integer value to the corresponding string representation of the
   component type.
   <BR> <BR>

   @param event       The event to request the component type from.
   @return            String representation of the component type.
   */
   public static String getComponentType(MelEntry event)
   {
      EventComponentType  type = event.getComponentType();
      if (type == EventComponentType.relative) {
         type = event.getComponentLocation().getComponentRelativeLocation().getComponentType();
      }

      for (int i = 0; i < m_ComponentTypes.length; i++)
      {
         if (m_ComponentTypeEnums[i] == type)
         {
             return m_ComponentTypes[i];
         }
      }    
      return m_ComponentTypes[0];   // This is the one for the Unknown component type
   }
      
   /**
   DESCRIPTION: <BR>
   Maps the event type to a string description of the event.
   <BR> <BR>

   @param event       The event to request the description from.
   @return            Textual description of the event.
   */
   public static String getDescription(MelEntry event)
   {
      String returnString = null;
      String eventString = EVENT_DESCRIPTION_KEY_PREFIX;
      eventString += getEventType(event);

      try
      {
         returnString = Resource.getString(eventString);
      }
      catch(Exception e)
      {
         // TODO Catch block may not be necessary, getString doesn't explicitly throw any exceptions
      }
      
      if ( returnString == null || returnString.length() == 0 )
      {
         return Resource.getString("EV_DESCRIPTION_UKNOWN");
      }
      else
      {         
         return returnString;
      }
   }
   
   /**
   DESCRIPTION: <BR>
   Used by Alert Manager for getting total description of the event. Any more details
   needed in the alert in future can be appended to the returned string.
   Maps the event type to a string description of the event and appends the component
   type and location to the string.
   <BR> <BR>

   @param event       The event to request the description from.
   @return            Textual description of the event.
   */
   public static String getTotalDescription(MelEntry event)
   {
      String returnString = null;
      String eventString = EVENT_DESCRIPTION_KEY_PREFIX;
      eventString += getEventType(event);

      returnString = Resource.getString(eventString);
      if ( returnString.length() == 0 )
      {
         return Resource.getString("EV_DESCRIPTION_UNKNOWN");
      }
      else
      {
         returnString += "\n";
         returnString += Resource.getString("EV_DETAIL_HEADER_COMP_TYPE");
         returnString += getComponentType(event);
         returnString += "\n";
         returnString += Resource.getString("EV_DETAIL_HEADER_COMP_LOCATION");
         returnString += getComponentLocation(event);
         return returnString;
      }
   }

   /**
   DESCRIPTION: <BR>
   Maps the event type to a US ASCII string description of the event.
   NOTE: java.util.Locale.US is a constant that creates a Locale object for _en_US. 
   The code will search for lsi.symsm.sam.EventResourceStrings_en_US and if it
   does not find this it will search for lsi.symsm.sam.EventResourceStrings_en
   and if it does not find this it will search for lsi.symsm.sam.EventResourceStrings.
   This will allow the default to be US, but also allow different defaults if the need arises 
   in the future.
   <BR> <BR>

   @param event       The event to request the description from.
   @return            Textual description of the event.
   */
   public static String getUSASCIIDescription(MelEntry event)
   {
      String returnString = null;
      String eventString = null;

      if (m_USASCIIResource == null)
      {
         m_USASCIIResource = ResourceBundle.getBundle("devmgr.versioned.sam.eventviewer.EventStrings",
                                                      Locale.US);
      }
       
      try
      {
         eventString = EVENT_DESCRIPTION_KEY_PREFIX;
         eventString += getEventType(event);
         returnString = m_USASCIIResource.getString(eventString);
      }
      catch (MissingResourceException e)
      { 
         // key was not found in resource class
         // return an empty string
         returnString = "";
      }

      if ( returnString.length() == 0 )
      {
         try
         {
            returnString = m_USASCIIResource.getString("EV_DESCRIPTION_UNKNOWN");
         }
         catch (MissingResourceException e)
         {    
             // key was not found in resource class
             // return an empty string
             returnString = "";
         }         
      }
      else
      {
            returnString += "\n\t";
            returnString += Resource.getString("EV_DETAIL_HEADER_COMP_TYPE");
            returnString += getComponentType(event);
            returnString += "\n\t";
            returnString += Resource.getString("EV_DETAIL_HEADER_COMP_LOCATION");
            returnString += getComponentLocation(event);    
            returnString += "\n\t";
            returnString += Resource.getString("EV_DETAIL_HEADER_PRIORITY");
            returnString += getEventPriorityString(event);
      }
      return returnString;
   }
   
   
   /**
   DESCRIPTION: <BR>
   Takes the sense key, ASC, and ASCQ fields in the melEntry and converts them
   to a string of the form XX/YY/ZZ. Where XX represents the sense key, YY represents
   the ASC and ZZ represents the ASCQ.
   <BR> <BR>

   @param event        The event to request the ASC/ASCQ from.
   @return             Sense Key/ASC/ASCQ
   */
   public static String getAscAscq(MelEntry event)
   {
      int sense = event.getSenseKey();
      int asc = event.getASC();
      int ascq = event.getASCQ();

      String all;
      // TODO: Fix this
      all = Integer.toHexString(sense) + 
      "/" +
      Integer.toHexString(asc) + 
      "/" +
      Integer.toHexString(ascq);
      return all;
   }

    /**
   DESCRIPTION: <BR>
   Converts Event source Controller in the melEntry to a string.
   <BR> <BR>

   @param event       The event to request the sequence number from.
   @return            Event source controller.
   */
   public static String getEventSourceController(MelEntry event)
   {  
       String returnString;

       int  controller = event.getEventSourceController();
       if ( controller == SLOT_A || controller == SLOT_B)
       {
          returnString = MessageFormat.format(Resource.getString("EV_COMPONENT_LOC_CONT"), 
                getControllerSlotName(controller));
       }
       else
       {
         returnString = Resource.getString("EV_COMPONENT_LOC_UNKNOWN");
       }
      return returnString;
   }
 
   /**
   Converts the mel component location or ExtendedComponentLocation to a string representation of 
   the component location.  returns "Not applicable" if location valid in the melEntry is set to
   false.
   If the Component type of the event is Controller, the code makes the assumption
   that slot 0 is Controller A and slot 1 is Controller B.
   If the Component type of the event is either Minihub or Channel, the code makes the
   assumption that slot 1 is Controller A and slot 2 is Controller B.

   @param event The event to request the component location from.
   @return A String representation of the location of the affected component.
    */
   public static String getComponentLocation(MelEntry event)
   {
      TraySlot traySlot = null;
      int                        tray;
      int                        groupNum;
      int                        channelPort;
      ChannelType                channelType;
      int                        slot;
      String          label;
      String                     returnString = Resource.getString("EV_COMPONENT_LOC_UNKNOWN");
      ComponentLocation location = event.getComponentLocation();
      EventComponentType         type = location.getEventComponentType();
      Object[]                   fourArguments = new Object[4];
      Object[]                   fiveArguments = new Object[5];
      ExtendedComponentLocation  extendedLocation = event.getExtComponentLocation();
      ExtendedComponentLocationType  extendedLocationType = event.getExtComponentLocation().getExtLocType();
      ComponentRelativeLocation  compRelLoc;
      ParentRelativeLocation [] parents;
      int traySlotId;
      EventComponentType componentType;
      int compLocation;
      int parentSize;
      
      // The ExtendedComponentLocation structure was added to provide more data for components
      // that previously had limited data. This structure should be checked before 
      // ComponentLocation, since it may have more information than ComponentLocation.
      if (extendedLocationType != ExtendedComponentLocationType.unknown)
      {
         switch (extendedLocationType)
         {
            case channel:
               returnString =
                  decodeChannelLocation(extendedLocation.getCompChannelLoc(), getEventType(event));
               break;
            
            case minihub:
               returnString = decodeMinihubLocation(extendedLocation.getCompMinihubLoc());
               break;
               
            case tray:
               returnString = decodeTray(extendedLocation.getCompTray());
               break;
            
            case trayslot:
               returnString = decodeTraySlot(extendedLocation.getCompTraySlot());
               break;
            
            case volume:
               label = location.getLabel();
               returnString = Resource.getString("EV_COMPONENT_VOLUME") + label;
               break;
            case pit:
               returnString = Resource.getString("EV_COMPONENT_PIT") +
                  extendedLocation.getPitSequence();
               break;
            default:
               //returnString is initialized to EV_COMPONENT_LOC_UNKNOWN
               break;               
         }
      }
      else
      {
         // The locationValid flag indicates whether the ComponentLocation data is valid.
         if (!event.isLocationValid())
         {
            return Resource.getString("EV_COMPONENT_LOC_UNKNOWN");
         }
         
         switch (type)
         {
            case unknown:
               returnString = Resource.getString("EV_COMPONENT_LOC_UNKNOWN");
               break;
   
            case drive:
               traySlot = location.getDriveLoc();
               returnString = decodeTraySlot(traySlot);
               break;
   
            case controller:
               traySlot = location.getControllerLoc();
               if (traySlot.getSlot() == 0) // Controller A
               {
                 returnString = MessageFormat.format(Resource.getString("EV_COMPONENT_LOC_CONT"), 
                        getControllerSlotName(SLOT_A));
               }
               else if (traySlot.getSlot() == 1) // Controller B
               {
                  returnString = MessageFormat.format(Resource.getString("EV_COMPONENT_LOC_CONT"), 
                        getControllerSlotName(SLOT_B));
               }
               break;
   
            case battery:
               tray = location.getBatteryTray();
               returnString = decodeTray(tray);
               break;
   
            case minihub:
               returnString = decodeMinihubLocation(location.getMinihubLoc());
               break;
   
            case channel:
               returnString = decodeChannelLocation(location.getChannelLoc(), getEventType(event));
               break;
   
            case powerSply:
            case esm:
            case fan:
              //Starting 9.12, firmware has put extended location information
              //for these three component type events, see PR84158
              traySlot = extendedLocation.getCompTraySlot();
              returnString = decodeTraySlot(traySlot);
              break;
   
            case enclosure:
               tray = location.getEnclosureTray();
               returnString = decodeTray(tray);
               break;
   
            case ups:
               returnString = Resource.getString("EV_COMPONENT_LOC_NA");
               break;
   
            case chip:
               traySlot = location.getChipLoc();
               returnString = decodeTraySlot(traySlot);
               break;
   
            case volume:
               label = location.getLabel();
               returnString = Resource.getString("EV_COMPONENT_VOLUME") + label;
               break;
   
            case volumeGrp:
               groupNum = location.getGroupNumber();
               returnString = Resource.getString("EV_COMPONENT_LOC_GROUP") + groupNum;
               break;
   
            case portCru:
               tray = location.getPortCruTray();
               returnString = decodeTray(tray);
               break;
   
            case tempSensor:
               tray = location.getTempSensorTray();
               returnString = decodeTray(tray);
               break;
               
            case interconnectCru:
               traySlot = location.getInterconnectCruLocation();
               returnString = decodeTraySlot(traySlot);
               break;
   
            case supportCru:
               traySlot = location.getSupportCruLocation();
               returnString = decodeTraySlot(traySlot);
               break;
   
            case alarm:
               traySlot = location.getAlarmLocation();
               returnString = decodeTraySlot(traySlot);
               break;
   
            case channelPort:
               channelPort = location.getChannelPort();
               returnString = Resource.getString("EV_COMPONENT_LOC_DRIVE_CHANNEL_PORT")
                               + channelPort;
               break;
               
            case sfpPort:
               traySlot = location.getSfpLocation();
               returnString = decodeTraySlot(traySlot);
               break;
   
            case hostBoard:
               returnString = Resource.getString("EV_COMPONENT_LOC_HOST_BOARD");
               break;

            case drawer:
               traySlot = location.getDrawerLocation();
               returnString = decodeTraySlot(traySlot);
               break;

              // added as part of CR~149350 due to a change in the MelEntry structure
            case relative:
              {
                 compRelLoc = event.getComponentLocation().getComponentRelativeLocation();
                 traySlotId = compRelLoc.getTrayId();
                 compLocation = compRelLoc.getRelativePosition();
                 componentType = compRelLoc.getComponentType();
                 parents = compRelLoc.getParent();
                 parentSize = compRelLoc.getParentSize();

//                 eventLogger.debug( "The component type is " + getComponentTypeString(componentType));

                 switch (componentType)
                 {
                    case drive:
                    case powerSply:
                    case fan:
                    case minihub:
                    case tempSensor:
                    case esm:
                    case controller:
                    case battery:
                    case enclosure:
                    case ups:
                    case interconnectCru:
                    case supportCru:
                    case alarm:
                    case sfpPort:
                    case hostBoard:
                    case ctlrSfp:
                    case ctlrSoc:
                    case hostIoCard:
                    case cacheBackupDevice:
                    case cacheMemDimm:
                    case hostPort:
                    case chip:
                    case channelPort:
                    case drawer:
                       returnString = decodeComponentRelativeLocation(parents, parentSize,
                                      traySlotId, componentType, compLocation);
                       break;
                    default:
                       //returnString is initialized to EV_COMPONENT_LOC_UNKNOWN
                       break;
                 }
              }
              break;

              //  FW will never give us this.  It is an internally used FW value only.
   
            case newFormat:
               returnString = Resource.getString("EV_COMPONENT_LOC_UNKNOWN");
               break;

            case ctlrSfp:
               channelType = location.getCtlrSfpLocation().getChannelType();
               slot        = location.getCtlrSfpLocation().getControllerSlot();
               fourArguments[0] = getControllerSlotName(slot);
               if ( channelType == ChannelType.hostside ) {
                  fourArguments[1] = Resource.getString("EV_CHANNEL_TYPE_HOST");
               }
               else if (channelType == ChannelType.driveside) {
                  fourArguments[1] = Resource.getString("EV_CHANNEL_TYPE_DRIVE");
               }
               fourArguments[2] = Integer.toString(location.getCtlrSfpLocation().getChannel());
               fourArguments[3] = Integer.toString(location.getCtlrSfpLocation().getSfpPort());
               returnString = MessageFormat.format(
                                 Resource.getString("EV_COMPONENT_LOC_CONTROLLER_SFP"),
                                 fourArguments );
                
               break;
   
            case ctlrSoc:
               channelType = location.getSocLocation().getChannelType();
               slot        = location.getSocLocation().getControllerSlot();
               fiveArguments[0] = getControllerSlotName(slot);
               if ( channelType == ChannelType.hostside ) {
                  fiveArguments[1] = Resource.getString("EV_CHANNEL_TYPE_HOST");
               }
               else if (channelType == ChannelType.driveside) {
                  fiveArguments[1] = Resource.getString("EV_CHANNEL_TYPE_DRIVE");
               }
               fiveArguments[2] = Integer.toString(location.getSocLocation().getChannel());
               fiveArguments[3] = Integer.toString(location.getSocLocation().getSoc());
               fiveArguments[4] = Integer.toString(location.getSocLocation().getPort());
               returnString = MessageFormat.format(
                                 Resource.getString("EV_COMPONENT_LOC_CONTROLLER_SOC"),
                                 fiveArguments );
               break;
               
            case initiator:
            	// TODO: Fix this up            	
//               returnString = Utilities.getStringFromUTF8EncodedBytes(
//                     location.getInitiatorLocation().getInitiatorName());
               break;
               
            case target:
// TODO: Fix this up            	
//               returnString = Utilities.getStringFromUTF8EncodedBytes(
//                     location.getTargetLocation().getTargetName());
               break;
               
            case isnsServer:
               returnString = MessageFormat.format(
                     Resource.getString("EV_COMPONENT_LOC_ISNS_SERVER"),
                     getStringForSYMbolIpAddress(
                           location.getIsnsServerLocation().getIpAddress()));
               break;
               
            case hostIoCard:
               returnString = MessageFormat.format(
                        Resource.getString("EV_COMPONENT_LOC_TRAY_SLOT"),
                        location.getHostIoCardLocation().getControllerTray(),
                        location.getHostIoCardLocation().getControllerSlot());
               break;
               
            case cacheBackupDevice:
               returnString = MessageFormat.format(
                     Resource.getString("EV_COMPONENT_LOC_CACHE_BACKUP_DEVICE"),
                     Integer.toString(
                           location.getCacheBackupDeviceLocation().getCacheBackupDeviceSlot()),
                           getControllerSlotName(
                                 location.getCacheBackupDeviceLocation().getControllerSlot()));
               break;
   
            case schedule:
               label = location.getSchedule();
               returnString = Resource.getString("EV_COMPONENT_VOLUME") + label;
               break;

            case asyncMirrorGroup:
               label = location.getAsyncMirrorGroup();
               returnString = Resource.getString("EV_COMPONENT_ASYNC_MIRROR_GROUP") + label;
               break;
   
            case diskPool:
               groupNum = location.getDiskPool();
               returnString = Resource.getString("EV_COMPONENT_DISK_POOL") + groupNum;
               break;
   
            case pit:
               label = location.getPitLabel();
               returnString = Resource.getString("EV_COMPONENT_PIT") + label;
               break;
   
            case pitConsistencyGroup:
               label = location.getCgLabel();
               returnString = Resource.getString("EV_COMPONENT_PIT_CONSISTENCY_GROUP") + label;
               break;
   
            case cgpit:
               label = location.getCgPitLabel();
               returnString = Resource.getString("EV_COMPONENT_CGPIT") + label;
               break;
   
            case cgview:
               label = location.getCgViewLabel();
               returnString = Resource.getString("EV_COMPONENT_CGVIEW") + label;
               break;
   
            case flashCache:
               returnString = Resource.getString("EV_COMPONENT_FLASH_CACHE");
               break;
   
            default:
               //returnString is initialized to EV_COMPONENT_LOC_UNKNOWN
               break;
         }
      }
      return returnString;
   }
   
   /**
    * A method to retun the Component Type String.
    * 
    * @param  compType
    * @return String representing the component
    */
   private static String getComponentTypeString(EventComponentType compType)
   {
      for (int i = 0; i < m_ComponentTypes.length; i++)
      {
         if (m_ComponentTypeEnums[i] == compType)
         {
             return m_ComponentTypes[i];
         }
      }    
      return m_ComponentTypes[0];  
   }

   /**
    * A method to display the location information of a component including it's parent's 
    * location.
    * 
    * @param parents - The array of parents of the component
    * @param traySlot - the tray Id of the component
    * @param compType - The component type
    * @param compLoc - the relative position of the component
    * @return
    */
   private static String decodeComponentRelativeLocation(ParentRelativeLocation [] parents, 
         int parentSize, int traySlot, EventComponentType compType, int compLoc)
   {
      StringBuffer location = new StringBuffer();
      String compLocation;
      
//      eventLogger.debug( "The parent size is  " + parentSize);

      // Case where the tray is a controller tray
      if ( traySlot == CONTROLLER_TRAY )
      {
//         eventLogger.debug( "Controller Tray");
         compLocation = Resource.getString("EV_COMPONENT_LOC_CONT_TRAY");
      }
      else
      {
         // if the tray is not a controller tray, obtain the array of parents and then loop 
         // through it and display the location in parent-child format.
         if (parentSize > 0)
         {
            location.append(Resource.getString("EV_COMPONENT_LOC_TRAY"));
            location.append(traySlot);
            for (int i = parentSize - 2; i >= 0; i--)
            {
               location.append(Resource.getString("COMMA_SPACE"));
               location.append(getComponentTypeString(parents[i].getComponentType()));
               location.append(Resource.getString("SPACE_CHARACTER"));
               location.append(parents[i].getRelativePosition());
            }

            location.append(Resource.getString("COMMA_SPACE"));
            location.append((compType != EventComponentType.drawer)?
            		Resource.getString("EV_COMPONENT_LOC_SLOT_ALT"):
            			Resource.getString("EV_DRAWER_ID"));
//            if (compType == EventComponentType.EVENT_COMP_TYPE_DRIVE)
//            {
//               location.append(JALUtil.getPositionName(compLoc, "DRIVE_LOCATION"));
//            } else
            if (compType == EventComponentType.controller)
            {
               location.append(getControllerSlotName(compLoc));
            }
            else
            {
               location.append(getComponentRelativePosition(compType, compLoc));
            }
            compLocation = location.toString();
         }
         else
         {
            location.append(Resource.getString("EV_COMPONENT_LOC_TRAY"));
            location.append(traySlot);

            if (compType != EventComponentType.enclosure)
            {
               location.append(Resource.getString("COMMA_SPACE"));
               location.append(Resource.getString("EV_COMPONENT_LOC_SLOT_ALT"));
               location.append(compLoc);
            }
            compLocation = location.toString();
         }
      }
      
//         eventLogger.debug( "The component location is  " + compLocation);
      return compLocation;
   }
   
   /**
   DESCRIPTION: <BR>
   Included here to eliminate duplicate code in switch statement in getComponentLocation().
   NOTE: This method returns a Tray and Slot location only (Tray X Slot Y) - it does not return
   the "Controller in Slot A/B" strings anymore.  In case the tray value is -1, then this
   method returns "Controller Tray".
   <BR><BR>

   @param traySlot   Tray and slot number to decode.
   @return           String to display in the detail and saved views.
   */
   public static String decodeTraySlot(TraySlot traySlot)
   {
      String      returnString = null;

      if ( traySlot.getTray() == CONTROLLER_TRAY )
      {
         returnString = Resource.getString("EV_COMPONENT_LOC_CONT_TRAY");
      }
      else if (traySlot.getTray() == TRAY_LOC_UNKNOWN)
      {
         returnString = Resource.getString("EV_COMPONENT_LOC_TRAY_UNKNOWN");
      }
      else
      {
         returnString = MessageFormat.format( Resource.getString("EV_COMPONENT_LOC_TRAY_SLOT"),
                        traySlot.getTray(), traySlot.getSlot());
      }
      return returnString;
   }
   
   /**
   DESCRIPTION: <BR>
   Checks for -1 in tray field an returns the appropriate string.  Included here to
   eliminate duplicate code in switch statement in getComponentLocation().
   <BR><BR>

   @param tray       Tray number to decode.
   @return           String to display in the detail and saved views.
   */
   public static String decodeTray(int tray)
   {
      String returnString = null;

      if ( tray == CONTROLLER_TRAY )
      {
         returnString = Resource.getString("EV_COMPONENT_LOC_CONT_TRAY");
      }
      else if (tray == TRAY_LOC_UNKNOWN)
      {
         returnString = Resource.getString("EV_COMPONENT_LOC_TRAY_UNKNOWN");
      }
      else
      {
         returnString = Resource.getString("EV_COMPONENT_LOC_TRAY") + tray;
      }

      return returnString;
   }
   
   /**
   DESCRIPTION: <BR>
   Converts the event type contained in the melEntry into a string.

   @param event       The event to request the event type from.
   @return            Event type as a 4 digit string(hex).  
   */
   public static String getEventType(MelEntry event)
   {
      String   hexString = null;
      int      type = event.getEventType();

      hexString = Integer.toHexString(type);
      return hexString.toUpperCase();
   }

   /**
   DESCRIPTION: <BR>
   Converts an integer value to the corresponding string representation of the
   event category.

   @param event   The event to request the category from.
   @return        String representation of the event category.
   */
   public static String getEventCategory(MelEntry event)
   {
      int   value = event.getCategory().value;

      if ( value < 0 || value >= m_Categories.length )
      {
         return m_Categories[0];
      }
      return m_Categories[value];
   }
   
   /**
   Decode a ChannelLocation object into a string that describes it. <P>
   
   NOTES/ASSUMPTIONS:  <BR>
   Ideally, this method would only require a ChannelLocation object to be passed in; however, there
   is a special case for channel miswire that must be tested, and it requires information about
   the event type. That is why the String representation of the event type is the second parameter.
   
   @param location The ChannelLocation object to be decoded into a textual representation
   @param eventType A String representation (in Hex) of the event type
   @return A String representation of this ChannelLocation
   */
   public static String decodeChannelLocation(ChannelLocation location, String eventType)
   {
      String   temp;
      String   channels = "";
      String   returnString = Resource.getString("EV_COMPONENT_LOC_UNKNOWN");
      ChannelType      side;
      int      slot;

      // the channel location is encoded. Each digit represents a channel number.
      temp = Integer.toString(location.getIndex());
      StringBuffer chBuf = new StringBuffer();
      if (temp.length() > 0)
      {
         chBuf.append(temp.substring(0, 1));
         for (int i=1; i < (temp.length()-1); i++)
         {
        	 chBuf.append(MessageFormat.format(
                        Resource.getString("EV_COMMA_SPACE_CHANNEL_NUMBER"),
                        temp.substring(i, i+1)));
         }
      }
      channels = chBuf.toString();
      side = location.getChannelType();
      
      if(eventType.equals("150F"))    // Channel Miswire
      {
         return MessageFormat.format(
                              Resource.getString("EV_COMPONENT_LOC_DRIVE_SIDE"), channels);
      }
      
      if(side == ChannelType.driveside)
      {
         returnString = MessageFormat.format(
                              Resource.getString("EV_COMPONENT_LOC_DRIVE_SIDE"), channels);
      }
      else if (side == ChannelType.hostside)
      {
         slot = location.getControllerSlot();
         returnString = MessageFormat.format(
            Resource.getString("EV_COMPONENT_LOC_HOST_SIDE"),
            getControllerSlotName(slot), channels);
      }
      
      return returnString;
   }
   
   /**
   Decode a MinihubLocation into a string that describes it. <P>
   
   NOTES/ASSUMPTIONS:  <BR>
   None.
   
   @param location The Minihublocation to be decoded into a textual
   @return A String representation of this MinihubLocation
   */
   public static String decodeMinihubLocation(MinihubLocation location)
   {
      String   returnString = Resource.getString("EV_COMPONENT_LOC_UNKNOWN");
      MinihubType      side;
      int      slot;
      
      side = location.getMinihubType();
      
      if (side == MinihubType.driveside)
      {
          returnString = MessageFormat.format(
                               Resource.getString("EV_COMPONENT_LOC_DRIVE_SIDE"),
                               location.getChannel());
      }
      else if (side == MinihubType.hostside)
      {
         slot = location.getControllerSlot();
         returnString = MessageFormat.format(
            Resource.getString("EV_COMPONENT_LOC_HOST_SIDE"),
            getControllerSlotName(slot),
            location.getChannel());
      }
      else
      {
         // otherwise, the SFP is on a drive tray and the controller slot
         // represents the SFP slot on that tray.
         returnString = decodeTray(location.getTray())
// TODO: Fix this up.         
//             + Resource.getGlobalString("SPECIAL_CHAR_COMMA")
             + Resource.getString("EV_COMPONENT_LOC_SFP_SLOT")
             + location.getControllerSlot();
      }


      return returnString;
   }
   
   public static String getStringForSYMbolIpAddress(IpVxAddress ipAddress)
   {
      String   result = IpAddressType.__UNDEFINED.toString();
      if (ipAddress.getAddressType() == IpAddressType.ipv4)
      {
          result=ipAddress.getIpv4Address();
      }
      else // IPV6
      {
          result=ipAddress.getIpv6Address();
      }
      return result;
   }
   
   /**
    * Translates the component's relative location w.r.t. it's parent's
    * location for Mercury. As Mercury has 0 based numbering for Components but
    * others uses 1 based numbering scheme.
    * 
    * @param componentType - Type of the component.
    * @param relativePosition - the relative of the component w.r.t. parent
    * @return Translated component's relative location.
    */
   private static int getComponentRelativePosition(EventComponentType componentType, int relativePosition)
   {
	  // TODO: Find a better solution.  This is the least disruptive.
      if (Resource.getString("OEM_SUPPORT").equals("MERCURY_SUPPORT"))
      {
         switch(componentType)
         {
            case controller:
            case battery:
            case drive:
               if (relativePosition != 0)
               {
                  return --relativePosition;
               }
               break;

            default:
               //Do nothing
               break;
         }
      }

      return relativePosition;
   }
    /**
     * Determines whether the MEL entry is alertable.
     * <P>
     * NOTES/ASSUMPTIONS: <BR>
     * None.
     *
     * @param entry MEL entry to check for alertability.
     * @return true if entry is alertable, false otherwise.
     */
    public static boolean isAlertable(MelEntry entry)
    {
        int override = entry.getControlParams().getOemSendAlert().value;
        boolean alertable;

        switch (entry.getControlParams().getOemSendAlert())
        {
            case acceptDefault:
                alertable =  entry.getControlParams().isDefaultSendAlert();
                break;

            case overrideToFalse:
                alertable = false;
                break;

            case overrideToTrue:
                alertable = true;
                break;

            default:
                assert false : override;
                // Don't know what the override is, use default
                alertable = entry.getControlParams().isDefaultSendAlert();
                break;
        }

        return alertable;
    }
    /**
     * Determines whether the MEL entry should trigger a needs attention condition.
     * <P>
     * NOTES/ASSUMPTIONS: <BR>
     * None.
     *
     * @param entry MEL entry to check for needs attention condition.
     * @return true if entry should trigger a needs attention condition, false otherwise.
     */
    public static boolean isNeedsAttention(MelEntry entry)
    {
        int override = entry.getControlParams().getOemNeedsAttention().value;
        boolean needsAttention;

        switch (entry.getControlParams().getOemNeedsAttention())
        {
            case acceptDefault:
                needsAttention =  entry.getControlParams().isDefaultNeedsAttention();
                break;

            case overrideToFalse:
                needsAttention = false;
                break;

            case overrideToTrue:
                needsAttention = true;
                break;

            default:
                assert false : override;
                // Don't know what the override is, use default
                needsAttention = entry.getControlParams().isDefaultNeedsAttention();
                break;
        }

        return needsAttention;
    }

}
