package devmgr.api.runtime;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.LinkedList;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * This class implements a decoder for XML input streams.  It provides
 * methods for processing such input data to construct Java objects
 * that coincide with the data in a stream.
 */
public class XmlDecoder
extends Decoder
{
    private static final int DEFAULT_MAX_STRING_LEN = 1024;
    
    private Document m_Doc;
    private Element m_CurrentElement;
    private LinkedList<DecoderContext> m_ContextList = 
        new LinkedList<DecoderContext>();
    
    /**
     * This class is used to track nested contexts that are needed
     * for structure and union processing.
     */
    private static class
    DecoderContext
    {
        ArrayList<Element> elements = null;
        int pos = 0;
    }
    
    /**
     * Constructor.
     * @param inStream Input stream with XML data.
     * @throws DecodeException
     */
    public XmlDecoder(InputStream inStream)
    throws DecodeException
    {
        try
        {
            DocumentBuilderFactory dbFactory = 
                DocumentBuilderFactory.newInstance();
            
            DocumentBuilder dBuilder = dbFactory.newDocumentBuilder();
            m_Doc = dBuilder.parse(inStream);
            m_Doc.getDocumentElement().normalize();
        }
        catch (Exception ex)
        {
            throw DecodeException.build(ex, "XML setup error");
        }

        m_CurrentElement = m_Doc.getDocumentElement();
    }
        
    /**
     * This method returns a list of all child elements of the
     * specified XML element.
     * @param elt The parent element
     * @return ArrayList of all child ELEMENTS of this parent.
     */
    private static ArrayList<Element>
    getChildElements(Element elt)
    {
        ArrayList<Element> list = new ArrayList<Element>();
        
        NodeList nodes = elt.getChildNodes();
        for (int i = 0; i < nodes.getLength(); i++)
        {
            Node node = nodes.item(i);
            if (node.getNodeType() == Node.ELEMENT_NODE)
                list.add((Element) node);
        }
        
        return list;
    }

    /**
     * This method retrieves the string that represents the text content
     * of the current element.
     * @param maxLen Maximum allowed length.
     * @return Current element's text content.
     * @throws DecodeException
     */
    private String
    fetchString(int maxLen)
    throws DecodeException
    {
        String s = m_CurrentElement.getTextContent();
        
        if (s.length() > maxLen)
            throw new DecodeException("XML: string too long");
        
        return s;
    }
    
    /**
     * This method retrieves the string that represents the text content
     * of the current element.
     * @return Current element's text content.
     * @throws DecodeException
     */
    private String
    fetchString()
    throws DecodeException
    {
        return fetchString(DEFAULT_MAX_STRING_LEN);
    }
    
    /**
     * This method retrieves a floating point value from the text content
     * of the current element.
     * @return (Double) floating point value.
     * @throws DecodeException
     */
    private double
    fetchDouble()
    throws DecodeException
    {
        String s = fetchString();
        double result = 0;
        
        try
        {
            result = Double.parseDouble(s);
        }
        catch (Exception ex)
        {
            throw DecodeException.build(ex, "XML: bad floating point");
        }
        
        return result;
    }
    
    /**
     * This method retrieves a long integer value from the text content
     * of the current element.
     * @return Long integer value.
     * @throws DecodeException
     */
    private long
    fetchInt()
    throws DecodeException
    {
        String s = fetchString();
        long result = 0;
        
        try
        {
            result = Long.parseLong(s);
        }
        catch (Exception ex)
        {
            throw DecodeException.build(ex, "XML: bad integer");
        }
        
        return result;
    }
    
    /**
     * This method gets the next "name" in the input stream.  This 
     * applies to both structure and union decoding. Note that in the
     * XML encoding scheme, the name is a found from the next element
     * in the input stream's current context. In addition to fetching
     * the name, we also set up the current XML element for consumption
     * by a follow-on decode operation that fetches its value.
     * @return Next name in the input stream, or null if end of object.
     * @throws DecodeException
     */
    private String 
    getName()
    throws DecodeException
    {
        DecoderContext decCtx = m_ContextList.getLast();
        
        // Check for end of context.
        if (decCtx.pos >= decCtx.elements.size())
            return null;
        
        // Consume the next element from the current context.
        m_CurrentElement = (Element) decCtx.elements.get(decCtx.pos++);
        
        // The element's node name is the "name" the caller wants.
        return m_CurrentElement.getNodeName();
    }

    /**
     * This method gets the next name (discriminator OR data field) in
     * the input stream for the specified union.  If we reach the end
     * of the union, a null value is returned.
     */
    @Override
    protected String 
    getUnionItem(String enclosingName, Context ctx, UnionSpec objType)
    throws DecodeException
    {
        return getName();
    }

    /**
     * This method gets the type binding information for the next
     * field in the input stream for the specified structure.  If we
     * reach the end of the structure, a null value is returned.
     */
    @Override
    protected TypeBinding
    getStructItem(String enclosingName, Context ctx, StructSpec objType)
    throws DecodeException
    {
        String name = getName();
        
        if (name == null)
            return null;

        TypeBinding binding = objType.nameMap.get(name);
        
        if (binding == null)
        {
            throw DecodeException.build
                ("XML: bad field name '%s' for class '%s'",
                 name, objType.cls.getCanonicalName());
        }
        
        return binding;
    }

    @Override
    protected byte decodeByte(String name)
    throws DecodeException
    {
        long result = fetchInt();
        return (byte) result;
    }

    @Override
    protected short decodeShort(String name)
    throws DecodeException
    {
        long result = fetchInt();
        return (short) result;
    }

    @Override
    protected int decodeInt(String name)
    throws DecodeException
    {
        long result = fetchInt();
        return (int) result;
    }

    @Override
    protected boolean
    isEndOfArray(String name, Context ctx)
    throws DecodeException
    {
        // In the XML encoding, if we're on the first element, then 
        // we are definitely NOT at the end of the array.  The fact that
        // there IS a first element means we have an item to process.
        if (ctx == Context.ARRAY_FIRST)
            return false;
        
        DecoderContext decCtx = m_ContextList.getLast();
        
        // If there are no more elements in the current context, we 
        // must be at the end of the array.
        if (decCtx.pos >= decCtx.elements.size())
            return true;
        
        // If the next element's name does NOT match the name of the
        // array, we must be at the end of the array.
        Element nextElt = (Element) decCtx.elements.get(decCtx.pos);
        if (! name.equals(nextElt.getNodeName()))
            return true;
        
        // Otherwise, we've got a new array element to process, so
        // make it ready for consumption.
        m_CurrentElement = nextElt;
        decCtx.pos++;
        
        return false;
    }

    @Override
    protected long decodeLong(String name)
    throws DecodeException
    {
        return fetchInt();
    }

    @Override
    protected boolean decodeBoolean(String name)
    throws DecodeException
    {
        String value = fetchString();
        
        if (value.equals("true"))
            return true;
        else if (value.equals("false"))
            return false;
        
        throw new DecodeException("XML: bad boolean");
    }

    @Override
    protected float decodeFloat(String name)
    throws DecodeException
    {
        return (float) fetchDouble();
    }

    @Override
    protected double decodeDouble(String name)
    throws DecodeException
    {
        return fetchDouble();
    }

    @Override
    protected String decodeString(String name, int maxLength)
    throws DecodeException
    {
        return fetchString(maxLength);
    }

    @Override
    protected String
    decodeOpaque(String name, OpaqueSpec typeSpec)
    throws DecodeException
    {
        String result = fetchString();
        
        if ((result.length() % 2) != 0)
            throw new DecodeException("XML: bad opaque data");
        
        if ( (! typeSpec.isVariable) && 
             ((result.length() / 2) != typeSpec.maxLength) )
            throw new DecodeException("XML: bad fixed opaque length");
        
        return result;
    }

    @Override
    protected void 
    decodeArrayStart(String name, Context ctx,
                     boolean isVariable, int maxLength)
    throws DecodeException
    {
    }

    @Override
    protected void 
    decodeArrayEnd(String name, boolean isVariable, int maxLength)
    throws DecodeException
    {
    }

    @Override
    protected boolean 
    decodeStructStart(String name, Context ctx, StructSpec spec)
    throws DecodeException
    {
        if (spec.isCollapsed)
            return true;
        
        // Create a new context for processing the structure.
        DecoderContext newCtx = new DecoderContext();
        newCtx.pos = 0;
        newCtx.elements = getChildElements(m_CurrentElement);
        
        m_ContextList.add(newCtx);
        return true;
    }

    @Override
    protected void 
    decodeStructEnd(String name, StructSpec spec)
    throws DecodeException
    {
        if (spec.isCollapsed)
            return;
        
        m_ContextList.removeLast();
    }

    @Override
    protected boolean 
    decodeUnionStart(String name, Context ctx, UnionSpec spec)
    throws DecodeException
    {
        // Create a new context for processing the union.
        DecoderContext newCtx = new DecoderContext();
        newCtx.pos = 0;
        newCtx.elements = getChildElements(m_CurrentElement);
        
        m_ContextList.add(newCtx);
        return true;
    }

    @Override
    protected void 
    decodeUnionEnd(String name, UnionSpec spec)
    throws DecodeException
    {
        m_ContextList.removeLast();
    }

    @Override
    protected EnumValue 
    decodeEnum(String name, EnumSpec typeSpec)
    throws DecodeException
    {
        String label = fetchString();
        
        Integer value = typeSpec.nameToValueMap.get(label);
        if (value == null)
            throw DecodeException.build
                    ("XML: bad enum value %s for %s in %s",
                     value, typeSpec.cls.getCanonicalName(), name);
        
        return new EnumValue(label, value);
    }
}