/**************************************************************************
 *                                                                         *
 *  Copyright 1998-2011 NetApp, Inc. All Rights Reserved.                  *
 *                                                                         *
 *  This file is confidential and a trade secret of NetApp, Inc.  The      *
 *  receipt of or possession of this file does not convey any rights to    *
 *  reproduce or disclose its contents or to manufacture, use, or sell     *
 *  anything it may describe, in whole, or in part, without the specific   *
 *  written consent of NetApp, Inc.                                        *
 *                                                                         *
 **************************************************************************/

package devmgr.api.security;

/**
 * VERSION: 3 <BR>
 * UPDATE DATE: Wed Mar 10 07:17:11 1999 <BR>
 * <BR>
 * CONTACTS: <BR>
 * Originator: Bill Delaney <BR>
 * Updater: bdelaney <BR>
 * <BR>
 * DESCRIPTION: <BR>
 * This class provides an implementation of the Data Encryption
 * Standard (DES) algorithm for encrypting and decrypting data
 * using 56-bit key strength.  It is based on the description of
 * the DES algorithm found in the following book:
 * <BR><BR>
 * "Applied Cryptography: Protocols, Algorithms and Source Code in C",
 * 2nd ed., by Bruce Schneier (John Wiley and Sons, 1996)
 * ISBN 0-471-12845-7.
 * <BR><BR>
 * NOTES/ASSUMPTIONS: <BR>
 * Software that implements encryption, particularly that
 * which uses the DES algorithm, is subject to stringent U.S.
 * export constraints.  Any product which uses this module must be
 * analyzed by the appropriate corporate entity to ensure that any
 * relevant export permits are obtained prior to general release.
 */
class DES
{
   /**
    * This constant value defines the size of each input/output
    * block.  The DES algorithm requires input blocks of exactly
    * this size, and produces equal sized output blocks for each
    * encryption/decryption operation.
    */
   public static final int BLOCK_SIZE = 8;

   /**
    * This array stores the internal subkeys that are used during
    * encryption/decryption operations.  Note that the only real
    * difference between these two processes is the order in which
    * the subkeys are stored when they are being generated using
    * the 56-bit DES key value.
    */
   private long[] m_Subkeys = new long[16];

   /**
    * This table contains the number of key rotations
    * performed during each of the 16 rounds of the DES encryption
    * algorithm.  In each round, each 28-bit half of the 56-bit
    * key is rotated left the specified number of bits.  The rotated
    * halves are first stored (logically) for further shifts in
    * successive rounds, and then fed through a key compression
    * permutation for explicit use during the round in question.
    */
   private static final byte[] keyRotations =
   {
      1, 1, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 2, 2, 2, 1
   };

   /**
    * This table contains the KEY SELECTION permutation bit
    * identifiers.  It is used to generate a 56-bit key from
    * a 64-bit input to the permutation function.
    *
    * Each index in the table corresponds to a bit position in the
    * output value of the permutation.  The value of a table entry
    * represents the bit position in the input value that generates
    * the indexed bit position in the output value of the permutation.
    * For example, the element at index 0 specifies the bit from the
    * input value that is to be used to determine bit 0 of the output
    * value.  Bit 0 always means the LEAST SIGNIFICANT BIT.
    */
   private static final byte[] keySelection =
   {
      60,   52,   44,   36,   59,   51,   43,   35,
      27,   19,   11,    3,   58,   50,   42,   34,
      26,   18,   10,    2,   57,   49,   41,   33,
      25,   17,    9,    1,   28,   20,   12,    4,
      61,   53,   45,   37,   29,   21,   13,    5,
      62,   54,   46,   38,   30,   22,   14,    6,
      63,   55,   47,   39,   31,   23,   15,    7
   };

   /**
    * This table contains the KEY COMPRESSION permutation
    * bit identifiers.  This table is used to generate a 48-bit
    * compressed key from a 56-bit input to the permutation function.
    *
    * Each index in the table corresponds to a bit position in the
    * output value of the permutation.  The value of a table entry
    * represents the bit position in the input value that generates
    * the indexed bit position in the output value of the permutation.
    * For example, the element at index 0 specifies the bit from the
    * input value that is to be used to determine bit 0 of the output
    * value.  Bit 0 always means the LEAST SIGNIFICANT BIT.
    */
   private static final byte[] keyCompression =
   {
      24,   27,   20,    6,   14,   10,    3,   22,
       0,   17,    7,   12,    8,   23,   11,    5,
      16,   26,    1,    9,   19,   25,    4,   15,
      54,   43,   36,   29,   49,   40,   48,   30,
      52,   44,   37,   33,   46,   35,   50,   41,
      28,   53,   51,   55,   32,   45,   39,   42
   };

   /**
    * This table contains the INITIAL PERMUTATION algorithm's
    * bit identifiers.  It is used to generate a 64-bit
    * permuted value from a 64-bit input to the permutation function.
    * The INITIAL PERMUTATION is performed on a 64-bit plaintext block
    * prior to performing the 16-round DES algorithm to encrypt the
    * block.
    *
    * Each index in the table corresponds to a bit position in the
    * output value of the permutation.  The value of a table entry
    * represents the bit position in the input value that generates
    * the indexed bit position in the output value of the permutation.
    * For example, the element at index 0 specifies the bit from the
    * input value that is to be used to determine bit 0 of the output
    * value.  Bit 0 always means the LEAST SIGNIFICANT BIT.
    */
   private static final byte[] initialPermutation =
   {
      57,   49,   41,   33,   25,   17,    9,    1,
      59,   51,   43,   35,   27,   19,   11,    3,
      61,   53,   45,   37,   29,   21,   13,    5,
      63,   55,   47,   39,   31,   23,   15,    7,
      56,   48,   40,   32,   24,   16,    8,    0,
      58,   50,   42,   34,   26,   18,   10,    2,
      60,   52,   44,   36,   28,   20,   12,    4,
      62,   54,   46,   38,   30,   22,   14,    6
   };

   /**
    * This table contains the FINAL PERMUTATION algorithm's
    * bit identifiers.  It is used to generate a 64-bit
    * permuted value from a 64-bit input to the permutation function.
    * The FINAL PERMUTATION is performed on the 64-bit output block
    * after the completion of the 16-round DES algorithm.  The result
    * is the 64-bit ciphertext block.
    *
    * Each index in the table corresponds to a bit position in the
    * output value of the permutation.  The value of a table entry
    * represents the bit position in the input value that generates
    * the indexed bit position in the output value of the permutation.
    * For example, the element at index 0 specifies the bit from the
    * input value that is to be used to determine bit 0 of the output
    * value.  Bit 0 always means the LEAST SIGNIFICANT BIT.
    */
   private static final byte[] finalPermutation =
   {
      39,    7,   47,   15,   55,   23,   63,   31,
      38,    6,   46,   14,   54,   22,   62,   30,
      37,    5,   45,   13,   53,   21,   61,   29,
      36,    4,   44,   12,   52,   20,   60,   28,
      35,    3,   43,   11,   51,   19,   59,   27,
      34,    2,   42,   10,   50,   18,   58,   26,
      33,    1,   41,    9,   49,   17,   57,   25,
      32,    0,   40,    8,   48,   16,   56,   24
   };

   /**
    * This table is used to implement both the S-box and P-box
    * permutations for the "S1" component of the DES encryption algorithm.
    * The S1-box, as described in the formal DES specification, operates
    * on the most significant 6 bits of a 48-bit intermediate value.  This
    * 6-bit value is fed into the S1-box, which produces a 4-bit result.
    * The 4 bits are then permuted out to fixed bit positions in the
    * 32-bit overall result of the S-box/P-box computation.  Here we use
    * one table to implement both aspects of the permutation.  The 6-bit
    * input value is used as an index into the table below.  The value
    * obtained from the table is then OR'ed with the corresponding values
    * obtained from the other 7 S-box/P-box table lookups, and the result
    * is used as the overall 32-bit result of the eight S-box/P-box
    * operations.
    */
   private static final int[] SP1 =
   {
      0x00808200, 0x00000000, 0x00008000, 0x00808202,
      0x00808002, 0x00008202, 0x00000002, 0x00008000,
      0x00000200, 0x00808200, 0x00808202, 0x00000200,
      0x00800202, 0x00808002, 0x00800000, 0x00000002,
      0x00000202, 0x00800200, 0x00800200, 0x00008200,
      0x00008200, 0x00808000, 0x00808000, 0x00800202,
      0x00008002, 0x00800002, 0x00800002, 0x00008002,
      0x00000000, 0x00000202, 0x00008202, 0x00800000,
      0x00008000, 0x00808202, 0x00000002, 0x00808000,
      0x00808200, 0x00800000, 0x00800000, 0x00000200,
      0x00808002, 0x00008000, 0x00008200, 0x00800002,
      0x00000200, 0x00000002, 0x00800202, 0x00008202,
      0x00808202, 0x00008002, 0x00808000, 0x00800202,
      0x00800002, 0x00000202, 0x00008202, 0x00808200,
      0x00000202, 0x00800200, 0x00800200, 0x00000000,
      0x00008002, 0x00008200, 0x00000000, 0x00808002
   };

   /**
    * This table is used to implement both the S-box and P-box
    * permutations for the "S2" component of the DES encryption algorithm.
    * See the description of the SP1 table for more details.
    */
   private static final int[] SP2 =
   {
      0x40084010, 0x40004000, 0x00004000, 0x00084010,
      0x00080000, 0x00000010, 0x40080010, 0x40004010,
      0x40000010, 0x40084010, 0x40084000, 0x40000000,
      0x40004000, 0x00080000, 0x00000010, 0x40080010,
      0x00084000, 0x00080010, 0x40004010, 0x00000000,
      0x40000000, 0x00004000, 0x00084010, 0x40080000,
      0x00080010, 0x40000010, 0x00000000, 0x00084000,
      0x00004010, 0x40084000, 0x40080000, 0x00004010,
      0x00000000, 0x00084010, 0x40080010, 0x00080000,
      0x40004010, 0x40080000, 0x40084000, 0x00004000,
      0x40080000, 0x40004000, 0x00000010, 0x40084010,
      0x00084010, 0x00000010, 0x00004000, 0x40000000,
      0x00004010, 0x40084000, 0x00080000, 0x40000010,
      0x00080010, 0x40004010, 0x40000010, 0x00080010,
      0x00084000, 0x00000000, 0x40004000, 0x00004010,
      0x40000000, 0x40080010, 0x40084010, 0x00084000
   };

   /**
    * This table is used to implement both the S-box and P-box
    * permutations for the "S3" component of the DES encryption algorithm.
    * See the description of the SP1 table for more details.
    */
   private static final int[] SP3 =
   {
      0x00000104, 0x04010100, 0x00000000, 0x04010004,
      0x04000100, 0x00000000, 0x00010104, 0x04000100,
      0x00010004, 0x04000004, 0x04000004, 0x00010000,
      0x04010104, 0x00010004, 0x04010000, 0x00000104,
      0x04000000, 0x00000004, 0x04010100, 0x00000100,
      0x00010100, 0x04010000, 0x04010004, 0x00010104,
      0x04000104, 0x00010100, 0x00010000, 0x04000104,
      0x00000004, 0x04010104, 0x00000100, 0x04000000,
      0x04010100, 0x04000000, 0x00010004, 0x00000104,
      0x00010000, 0x04010100, 0x04000100, 0x00000000,
      0x00000100, 0x00010004, 0x04010104, 0x04000100,
      0x04000004, 0x00000100, 0x00000000, 0x04010004,
      0x04000104, 0x00010000, 0x04000000, 0x04010104,
      0x00000004, 0x00010104, 0x00010100, 0x04000004,
      0x04010000, 0x04000104, 0x00000104, 0x04010000,
      0x00010104, 0x00000004, 0x04010004, 0x00010100
   };

   /**
    * This table is used to implement both the S-box and P-box
    * permutations for the "S4" component of the DES encryption algorithm.
    * See the description of the SP1 table for more details.
    */
   private static final int[] SP4 =
   {
      0x80401000, 0x80001040, 0x80001040, 0x00000040,
      0x00401040, 0x80400040, 0x80400000, 0x80001000,
      0x00000000, 0x00401000, 0x00401000, 0x80401040,
      0x80000040, 0x00000000, 0x00400040, 0x80400000,
      0x80000000, 0x00001000, 0x00400000, 0x80401000,
      0x00000040, 0x00400000, 0x80001000, 0x00001040,
      0x80400040, 0x80000000, 0x00001040, 0x00400040,
      0x00001000, 0x00401040, 0x80401040, 0x80000040,
      0x00400040, 0x80400000, 0x00401000, 0x80401040,
      0x80000040, 0x00000000, 0x00000000, 0x00401000,
      0x00001040, 0x00400040, 0x80400040, 0x80000000,
      0x80401000, 0x80001040, 0x80001040, 0x00000040,
      0x80401040, 0x80000040, 0x80000000, 0x00001000,
      0x80400000, 0x80001000, 0x00401040, 0x80400040,
      0x80001000, 0x00001040, 0x00400000, 0x80401000,
      0x00000040, 0x00400000, 0x00001000, 0x00401040
   };

   /**
    * This table is used to implement both the S-box and P-box
    * permutations for the "S5" component of the DES encryption algorithm.
    * See the description of the SP1 table for more details.
    */
   private static final int[] SP5 =
   {
      0x00000080, 0x01040080, 0x01040000, 0x21000080,
      0x00040000, 0x00000080, 0x20000000, 0x01040000,
      0x20040080, 0x00040000, 0x01000080, 0x20040080,
      0x21000080, 0x21040000, 0x00040080, 0x20000000,
      0x01000000, 0x20040000, 0x20040000, 0x00000000,
      0x20000080, 0x21040080, 0x21040080, 0x01000080,
      0x21040000, 0x20000080, 0x00000000, 0x21000000,
      0x01040080, 0x01000000, 0x21000000, 0x00040080,
      0x00040000, 0x21000080, 0x00000080, 0x01000000,
      0x20000000, 0x01040000, 0x21000080, 0x20040080,
      0x01000080, 0x20000000, 0x21040000, 0x01040080,
      0x20040080, 0x00000080, 0x01000000, 0x21040000,
      0x21040080, 0x00040080, 0x21000000, 0x21040080,
      0x01040000, 0x00000000, 0x20040000, 0x21000000,
      0x00040080, 0x01000080, 0x20000080, 0x00040000,
      0x00000000, 0x20040000, 0x01040080, 0x20000080
   };

   /**
    * This table is used to implement both the S-box and P-box
    * permutations for the "S6" component of the DES encryption algorithm.
    * See the description of the SP1 table for more details.
    */
   private static final int[] SP6 =
   {
      0x10000008, 0x10200000, 0x00002000, 0x10202008,
      0x10200000, 0x00000008, 0x10202008, 0x00200000,
      0x10002000, 0x00202008, 0x00200000, 0x10000008,
      0x00200008, 0x10002000, 0x10000000, 0x00002008,
      0x00000000, 0x00200008, 0x10002008, 0x00002000,
      0x00202000, 0x10002008, 0x00000008, 0x10200008,
      0x10200008, 0x00000000, 0x00202008, 0x10202000,
      0x00002008, 0x00202000, 0x10202000, 0x10000000,
      0x10002000, 0x00000008, 0x10200008, 0x00202000,
      0x10202008, 0x00200000, 0x00002008, 0x10000008,
      0x00200000, 0x10002000, 0x10000000, 0x00002008,
      0x10000008, 0x10202008, 0x00202000, 0x10200000,
      0x00202008, 0x10202000, 0x00000000, 0x10200008,
      0x00000008, 0x00002000, 0x10200000, 0x00202008,
      0x00002000, 0x00200008, 0x10002008, 0x00000000,
      0x10202000, 0x10000000, 0x00200008, 0x10002008
   };

   /**
    * This table is used to implement both the S-box and P-box
    * permutations for the "S7" component of the DES encryption algorithm.
    * See the description of the SP1 table for more details.
    */
   private static final int[] SP7 =
   {
      0x00100000, 0x02100001, 0x02000401, 0x00000000,
      0x00000400, 0x02000401, 0x00100401, 0x02100400,
      0x02100401, 0x00100000, 0x00000000, 0x02000001,
      0x00000001, 0x02000000, 0x02100001, 0x00000401,
      0x02000400, 0x00100401, 0x00100001, 0x02000400,
      0x02000001, 0x02100000, 0x02100400, 0x00100001,
      0x02100000, 0x00000400, 0x00000401, 0x02100401,
      0x00100400, 0x00000001, 0x02000000, 0x00100400,
      0x02000000, 0x00100400, 0x00100000, 0x02000401,
      0x02000401, 0x02100001, 0x02100001, 0x00000001,
      0x00100001, 0x02000000, 0x02000400, 0x00100000,
      0x02100400, 0x00000401, 0x00100401, 0x02100400,
      0x00000401, 0x02000001, 0x02100401, 0x02100000,
      0x00100400, 0x00000000, 0x00000001, 0x02100401,
      0x00000000, 0x00100401, 0x02100000, 0x00000400,
      0x02000001, 0x02000400, 0x00000400, 0x00100001
   };

   /**
    * This table is used to implement both the S-box and P-box
    * permutations for the "S8" component of the DES encryption algorithm.
    * See the description of the SP1 table for more details.
    */
   private static final int[] SP8 =
   {
      0x08000820, 0x00000800, 0x00020000, 0x08020820,
      0x08000000, 0x08000820, 0x00000020, 0x08000000,
      0x00020020, 0x08020000, 0x08020820, 0x00020800,
      0x08020800, 0x00020820, 0x00000800, 0x00000020,
      0x08020000, 0x08000020, 0x08000800, 0x00000820,
      0x00020800, 0x00020020, 0x08020020, 0x08020800,
      0x00000820, 0x00000000, 0x00000000, 0x08020020,
      0x08000020, 0x08000800, 0x00020820, 0x00020000,
      0x00020820, 0x00020000, 0x08020800, 0x00000800,
      0x00000020, 0x08020020, 0x00000800, 0x00020820,
      0x08000800, 0x00000020, 0x08000020, 0x08020000,
      0x08020020, 0x08000000, 0x00020000, 0x08000820,
      0x00000000, 0x08020820, 0x00020020, 0x08000020,
      0x08020000, 0x08000800, 0x08000820, 0x00000000,
      0x08020820, 0x00020800, 0x00020800, 0x00000820,
      0x00000820, 0x00020020, 0x08000000, 0x08020800
   };

   /**
    * DESCRIPTION: <BR>
    * This method computes a table-based permutation on the
    * input value, returning a 64-bit output result.  The caller
    * must specify the permutation table to be used for the
    * computation.  The number of valid bits in the output value
    * is determined by the number of elements in the table.
    * <BR><BR>
    * NOTES/ASSUMPTIONS: <BR>
    * None.
    * <BR>
    * @param input The input value
    * @param table The table that defines the permutation
    */
   private long permute(long input,
                        byte[] table)
   {
      int numBits = table.length;
      long output = 0;

      /*
       * Construct each bit of the output value, starting with the
       * LEAST SIGNIFICANT BIT.  Do this by using the permutation
       * table to determine the input bit that corresponds to each
       * successive output bit position, then OR-ing in the appropriate
       * shifted input bit value.
       */
      for (int i = 0; i < numBits; i++)
      {
         if ( (input & (1L << table[i])) != 0 )
         {
            output |= (1L << i);
         }
      }

      return output;
   }

   /**
    * DESCRIPTION: <BR>
    * This method loads a 64-bit value from memory.  The first
    * byte (i.e. at the lowest memory address) corresponds to the
    * MOST SIGNIFICANT BYTE of the 64-bit value.
    * <BR><BR>
    * NOTES/ASSUMPTIONS: <BR>
    * None.
    * <BR>
    * @param buf The memory buffer with a 64-bit (i.e. 8-byte) value
    */
   protected long load64(byte[] buf)
   {
      long result;

      result = ( ((long)(buf[0]) & 0xFF) << 56 ) |
               ( ((long)(buf[1]) & 0xFF) << 48 ) |
               ( ((long)(buf[2]) & 0xFF) << 40 ) |
               ( ((long)(buf[3]) & 0xFF) << 32 ) |
               ( ((long)(buf[4]) & 0xFF) << 24 ) |
               ( ((long)(buf[5]) & 0xFF) << 16 ) |
               ( ((long)(buf[6]) & 0xFF) <<  8 ) |
               ( ((long)(buf[7]) & 0xFF)       );

      return result;
   }

   /**
    * DESCRIPTION: <BR>
    * This method stores a 64-bit value into memory.  The first
    * byte (i.e. at the lowest memory address) corresponds to the
    * MOST SIGNIFICANT BYTE of the 64-bit value.
    * <BR><BR>
    * NOTES/ASSUMPTIONS: <BR>
    * None.
    * <BR>
    * @param value The value to be stored in memory
    * @param buf The memory buffer for the 64-bit (i.e. 8-byte) value
    */
   protected void store64(long value, byte[] buf)
   {
      int i = 0;

      buf[i++] = (byte)(value >>> 56);
      buf[i++] = (byte)(value >>> 48);
      buf[i++] = (byte)(value >>> 40);
      buf[i++] = (byte)(value >>> 32);
      buf[i++] = (byte)(value >>> 24);
      buf[i++] = (byte)(value >>> 16);
      buf[i++] = (byte)(value >>>  8);
      buf[i  ] = (byte)(value      );

      return;
   }

   /**
    * DESCRIPTION: <BR>
    * This method performs a right-rotation of the specified 32-bit
    * operand, and returns the resulting value.
    * <BR><BR>
    * NOTES/ASSUMPTIONS: <BR>
    * None.
    * <BR>
    * @param value The value to be rotated.
    * @param bits The number of bit positions to rotate.
    */
   private int rotateRight(int value, int bits)
   {
      return ( (value >>> bits) | (value << (32 - bits)) );
   }

   /**
    * DESCRIPTION: <BR>
    * This method performs a left-rotation of the specified 32-bit
    * operand, and returns the resulting value.
    * <BR><BR>
    * NOTES/ASSUMPTIONS: <BR>
    * None.
    * <BR>
    * @param value The value to be rotated.
    * @param bits The number of bit positions to rotate.
    */
   private int rotateLeft(int value, int bits)
   {
      return ( (value << bits) | (value >>> (32 - bits)) );
   }

   /**
    * DESCRIPTION: <BR>
    * This method initializes the DES object for either
    * encryption or decryption operations using the given 64-bit
    * key value.  The key value itself is used to compute a series
    * of 16 different 48-bit "subkeys", as defined in the formal DES
    * specification.  Each subkey ends up being used in one of the
    * 16 rounds of the encryption algorithm.
    * <BR><BR>
    * NOTES/ASSUMPTIONS: <BR>
    * None.
    * <BR>
    * @param key Buffer with 8-byte key data
    * @param encrypt True means encrypt, false means decrypt
    */
   public void setKey(byte[] key, boolean encrypt)
   {
      long p;
      long left;
      long right;
      int bits;
      int index;
      int incr;

      /*
       * Determine the index values to use when filling in the subkey
       * table in our instance data.  For encryption, we just fill
       * it in starting with index 0 and moving forward.  For decryption,
       * we start at the last index and move in reverse.  The ability to
       * perform encryption or decryption using the same process, but
       * with reversed subkey indexes, is part of the "magic" of the DES
       * algorithm.
       */
      if (encrypt)
      {
         index = 0;
         incr = 1;
      }
      else
      {
         index = 15;
         incr = -1;
      }

      /*
       * Permute the input block to obtain a 56-bit key value that will
       * be used as the basis for our subkeys.  This bit selection is
       * rigidly specified by the DES standard.
       */
      p = permute(load64(key), keySelection);

      /*
       * Split the initial 56-bit value into two 28-bit halves that will
       * be manipulated independently in computing subkeys.  Note that
       * we don't do any shifting here, we just isolate the appropriate
       * bits in each of the variables.
       */
      left = p & ~0x0FFFFFFF;
      right = p & 0x0FFFFFFF;

      /*
       * The following loop computes one subkey per iteration, for a total
       * of 16 different subkeys, as required by the DES algorithm.  The
       * subkey is computed by rotating each half (circularly) a prescribed
       * number of bits per iteration.  The shifted halves are merged into
       * a 56-bit value that is fed into a compression permutation to yield
       * the desired 48-bit subkey.
       */
      for (int i = 0; i < 16; i++)
      {
         bits = keyRotations[i];

         left = ( (left << bits) | (left >>> (28 - bits)) ) & ~0x0FFFFFFF;
         right = ( (right << bits) | (right >>> (28 - bits)) ) & 0x0FFFFFFF;

         m_Subkeys[index] = permute(left | right, keyCompression);
         index += incr;
      }

      return;
   }

   /**
    * DESCRIPTION: <BR>
    * Perform the DES algorithm for either encryption or decryption
    * of the specified 8-byte input block.  The choice between
    * encryption and decryption is made when the key value is
    * established for the object.  This algorithm operates the
    * same way, regardless of the direction.
    * <BR><BR>
    * NOTES/ASSUMPTIONS: <BR>
    * None.
    * <BR>
    * @param input The 8-byte (64-bit) input data block
    * @param output The 8-byte (64-bit) output data block
    */
   public void crypt(byte[] input, byte[] output)
   {
      /*
       * Just load the input as a 64-bit value and invoke the
       * method that operates on such entities, storing the
       * result back into the caller's output buffer.
       */
      store64(crypt(load64(input)), output);

      return;
   }

   /**
    * DESCRIPTION: <BR>
    * Perform the DES algorithm for either encryption or decryption
    * of the specified 64-bit input block.  The choice between
    * encryption and decryption is made when the key value is
    * established for the object.  This algorithm operates the
    * same way, regardless of the direction.
    * <BR><BR>
    * NOTES/ASSUMPTIONS: <BR>
    * None.
    * <BR>
    * @param input The 64-bit input data value
    * @return The 64-bit output data value
    */
   protected long crypt(long input)
   {
      int i;
      int left;
      int right;
      int w;
      int f;
      int rNew;
      long K;
      long p;

      /*
       * Perform the initial permutation on the data, as prescribed by the
       * DES algorithm.  This is a straight 64-bit to 64-bit permutation.
       */
      p = permute(input, initialPermutation);

      /*
       * Split the 64-bit value into two 32-bit halves that will be
       * maintained during the course of the 16-round computation.
       */
      left = (int)(p >>> 32);
      right = (int)(p);

      /*
       * The following loop implements the 16-round DES computation.
       * We assemble the output value of the DES "f" function by feeding
       * 6-bit values into each of the eight combined S-box/P-box tables,
       * and OR-ing the result values together.  This output is then XOR-ed
       * with the input left-side value to produce the new right-side value.
       * The new left-side value for each iteration is just a pass-thru of
       * the initial right-side value.
       *
       * To determine the 6-bit input value for the S-box/P-box operations,
       * we use the expansion-permutation output from the initial right-side
       * value, XOR-ed with 6-bit chunks of the subkey for the round.  We
       * achieve the 4-bit to 6-bit expansion by first rotating the right-side
       * value left by one bit to set the initial bit positions, and then
       * iteratively rotating right by four bits and selecting the least
       * significant six bits of the rotated result.
       */
      for (i = 0; i < 16; i++)
      {
         K = m_Subkeys[i];

         w = rotateLeft(right, 1);
         f = SP8[(w ^ (int)(K >>> 0)) & 0x3F];

         w = rotateRight(w, 4);
         f |= SP7[(w ^ (int)(K >>>  6)) & 0x3F];

         w = rotateRight(w, 4);
         f |= SP6[(w ^ (int)(K >>> 12)) & 0x3F];

         w = rotateRight(w, 4);
         f |= SP5[(w ^ (int)(K >>> 18)) & 0x3F];

         w = rotateRight(w, 4);
         f |= SP4[(w ^ (int)(K >>> 24)) & 0x3F];

         w = rotateRight(w, 4);
         f |= SP3[(w ^ (int)(K >>> 30)) & 0x3F];

         w = rotateRight(w, 4);
         f |= SP2[(w ^ (int)(K >>> 36)) & 0x3F];

         w = rotateRight(w, 4);
         f |= SP1[(w ^ (int)(K >>> 42)) & 0x3F];

         rNew = left ^ f;
         left = right;
         right = rNew;
      }

      /*
       * Merge the left and right sides together to form a 64-bit result,
       * then feed it into the straight (bit preserving) FINAL PERMUTATION
       * to determine the overall result.  Note that our loop above swaps
       * the left and (permuted) right sides after EACH round.  However, DES
       * specifies that the swap be skipped after the last round, so we
       * reverse that "errant" final swap by swapping again here when we
       * are merging the halves into a 64-bit result.
       */
      p = ((long)(right) << 32) | ((long)(left) & 0xFFFFFFFFL);
      p = permute(p, finalPermutation);

      /*
       * Return the output value to the caller.
       */
      return p;
   }
}
