/**************************************************************************
 *                                                                         *
 *  Copyright 1998-2011 NetApp, Inc. All Rights Reserved.                  *
 *                                                                         *
 *  This file is confidential and a trade secret of NetApp, Inc.  The      *
 *  receipt of or possession of this file does not convey any rights to    *
 *  reproduce or disclose its contents or to manufacture, use, or sell     *
 *  anything it may describe, in whole, or in part, without the specific   *
 *  written consent of NetApp, Inc.                                        *
 *                                                                         *
 **************************************************************************/

package devmgr.api.security;

/**
 * VERSION: 3 <BR>
 * UPDATE DATE: Wed Mar 10 07:17:14 1999 <BR>
 * <BR>
 * CONTACTS: <BR>
 * Originator: Bill Delaney <BR>
 * Updater: bdelaney <BR>
 * <BR>
 * DESCRIPTION: <BR>
 * This class adds Cipher Block Chaining capability to the DES
 * class by extending it and then handling all required block
 * chaining logic.
 * <BR><BR>
 * NOTES/ASSUMPTIONS: <BR>
 * None.
 */
class DESWithCBC extends DES
{
   private long m_ChainValue;
   private boolean m_Encrypting;

   /**
    * DESCRIPTION: <BR>
    * This constructor initializes the object for further encryption or
    * decryption operations using the given 64-bit key value and
    * initialization vector.
    * <BR><BR>
    * NOTES/ASSUMPTIONS: <BR>
    * None.
    * <BR>
    * @param key Buffer with 8-byte key data
    * @param encrypt True means encrypt, false means decrypt
    * @param IV The initialization vector for block chaining
    */
   public DESWithCBC(byte[] key, boolean encrypt, byte[] IV)
   {
      m_ChainValue = load64(IV);
      m_Encrypting = encrypt;

      super.setKey(key, encrypt);
   }

   /**
    * DESCRIPTION: <BR>
    * Perform the DES algorithm for either encryption or decryption
    * of the specified 8-byte input block, accounting for cipher
    * block chaining in the process.
    * <BR><BR>
    * NOTES/ASSUMPTIONS: <BR>
    * None.
    * <BR>
    * @param input The 8-byte (64-bit) input data block
    * @param output The 8-byte (64-bit) output data block
    */
   public void crypt(byte[] input, byte[] output)
   {
      long inValue = load64(input);
      long outValue;

      /*
       * If we're encrypting, apply the current chaining value to the
       * input data prior to feeding it to the actual encryption engine.
       */
      if (m_Encrypting)
      {
         inValue ^= m_ChainValue;
      }

      /*
       * Perform the actual encryption/decryption operation on the
       * input block here by invoking the superclass to do it for us.
       */
      outValue = super.crypt(inValue);

      /*
       * Handle chaining again here.  If we're encrypting, save off the
       * ciphertext as the chain value to apply to the next input block.
       * If we're decrypting, apply the saved chaining value to the
       * decrypted block, then save the original input value as the
       * chaining value for the next block to be processed.
       */
      if (m_Encrypting)
      {
         m_ChainValue = outValue;
      }
      else
      {
         outValue ^= m_ChainValue;
         m_ChainValue = inValue;
      }

      /*
       * Store the result into the caller's buffer.
       */
      store64(outValue, output);
   }
}
