/******************************************************************************
 * Copyright 2015 NetApp, Inc. All rights reserved.                           *
 *                                                                            *
 * This file is confidential and a trade secret of NetApp, Inc.               *
 * The receipt of or possession of this file does not convey any rights to    *
 * reproduce or disclose its contents or to manufacture, use, or sell         *
 * anything it may describe, in whole, or in part, without the specific       *
 * written consent of NetApp, Inc.                                            *
 ******************************************************************************/

package devmgr.api.security;

import javax.crypto.Cipher;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;
import java.security.GeneralSecurityException;
import java.security.MessageDigest;

/**
 * PBE implementation using a MessageDigest and Cipher.
 * <P>
 * The JRE supports PBE via the Cipher class and using an algorithm name such as
 * PBEWith[Digest]And[Encryption].  Ideally we would use those.  However, we are using SHA-256
 * as the digest which none of our supported JVMs support out of the box.  Additionally, we
 * use an initialization vector based on the output of the key derivation function which doesn't
 * seem to be directly supported by the JRE's PBE algorithm so we would have to calculate that
 * anyway.
 * <P>
 * REFERENCE: <BR>
 * http://ictwiki.eng.netapp.com/index.php/SYMbol_security_enhancement <BR>
 * https://www.ietf.org/rfc/rfc2898.txt <BR>
 * http://csrc.nist.gov/publications/nistpubs/800-131A/sp800-131A.pdf
 */
class PbeWithDigestAndCipher {
    private final PbeType pbeType;
    private Cipher cipher;


    /**
     * Creates a new PBEWithDigestAndCipher.
     *
     * @param pbeType type of PBE to use
     */
    public PbeWithDigestAndCipher(PbeType pbeType) {
        this.pbeType = pbeType;
    }


    /**
     * Initializes the PBE object for all ensuing operations.
     * <P>
     * This must be called before each time the crypt method is called.
     *
     * @param encrypt True to encrypt, false to decrypt
     * @param password Buffer with password data (raw bytes)
     * @param salt salt value for generating key and IV
     * @param iterationCount number of iterations for generating key and IV
     * @throws GeneralSecurityException if an error occurs initializing this PBE
     */
    public void init(boolean encrypt, byte[] password, byte[] salt, int iterationCount)
            throws GeneralSecurityException {
        assert salt.length == pbeType.saltSize;

        MessageDigest digest = MessageDigest.getInstance(pbeType.digestName);
        digest.update(password);
        digest.update(salt, 0, pbeType.saltSize);

        byte[] digestBytes;
        for (int i = 1; i < iterationCount; i++) {
            digestBytes = digest.digest();
            digest.update(digestBytes);
        }
        digestBytes = digest.digest();

        SecretKeySpec encryptionKeySpec = new SecretKeySpec(digestBytes, 0, pbeType.keySize,
                pbeType.cipherKeyName);
        IvParameterSpec ivSpec = new IvParameterSpec(digestBytes,
                digestBytes.length - pbeType.ivSize, pbeType.ivSize);

        cipher = Cipher.getInstance(pbeType.cipherName);
        cipher.init(encrypt ? Cipher.ENCRYPT_MODE : Cipher.DECRYPT_MODE, encryptionKeySpec, ivSpec);
    }


    /**
     * Encrypts or decrypts the input.
     * <P>
     * This can only be called once for each init call.
     *
     * @param input data to either encrypt or decrpyt based on the previous init call
     * @return encrypted or decrypted data
     * @throws GeneralSecurityException if an error occurs encrypting/decrypting the input
     */
    public byte[] crypt(byte[] input) throws GeneralSecurityException {
        return cipher.doFinal(input);
    }


    /**
     * Gets the size of the salt to use.
     *
     * @return size of salt in bytes
     */
    public int getSaltSize() { return pbeType.saltSize; }


    /**
     * Parameters for PBE.
     */
    public enum PbeType
    {
        /**
         * Original SYMbol PBE type
         */
        PKCS5_SHA1_DES("SHA-1", "DES/CBC/PKCS5Padding", "DES", 8, 8, 8),

        /**
         * Updated SYMbol PBE type to conform with SP800-131A
         */
        PKCS5_SHA256_AES("SHA-256", "AES/CBC/PKCS5Padding", "AES", 8, 16, 16);


        private final String digestName;
        private final String cipherName;
        private final String cipherKeyName;
        private final int saltSize;
        private final int keySize;
        private final int ivSize;


        /**
         * Creates a new PBEType.
         *
         * @param digestName MessageDigest name of hashing function
         * @param cipherName Cipher name of encryption type
         * @param cipherKeyName Cipher name to use for Key
         * @param saltSize size of salt in bytes
         * @param keySize size of encryption key in bytes
         * @param ivSize size of encryption initialization vector in bytes
         */
        PbeType(String digestName, String cipherName, String cipherKeyName, int saltSize,
                int keySize, int ivSize) {
            this.digestName = digestName;
            this.cipherName = cipherName;
            this.cipherKeyName = cipherKeyName;
            this.saltSize = saltSize;
            this.keySize = keySize;
            this.ivSize = ivSize;
        }
    }
}
